import datetime
import pathlib
import warnings
from typing import Optional

import typer
from rich.console import Console
from rich.json import JSON
from rich.table import Table

from actual import Actual, get_accounts, get_transactions
from actual.cli.config import BudgetConfig, Config, OutputType, State
from actual.queries import get_payees
from actual.version import __version__

# avoid displaying warnings on a CLI
warnings.filterwarnings("ignore")

app = typer.Typer()

console = Console()
config: Config = Config.load()
state: State = State()


@app.callback()
def main(output: OutputType = typer.Option("table", "--output", "-o", help="Output format: table or json")):
    if output:
        state.output = output


@app.command()
def init(
    url: str = typer.Option(None, "--url", help="URL of the actual server"),
    password: str = typer.Option(None, "--password", help="Password for the budget"),
    encryption_password: str = typer.Option(None, "--encryption-password", help="Encryption password for the budget"),
    context: str = typer.Option(None, "--context", help="Context for this budget context"),
    file_id: str = typer.Option(None, "--file", help="File ID or name on the remote server"),
):
    """
    Initializes an actual budget config interactively if options are not provided.
    """
    if not url:
        url = typer.prompt("Please enter the URL of the actual server", default="http://localhost:5006")

    if not password:
        password = typer.prompt("Please enter the Actual server password", hide_input=True)

    # test the login
    server = Actual(url, password=password)

    if not file_id:
        files = server.list_user_files()
        options = [file for file in files.data if not file.deleted]
        for idx, option in enumerate(options):
            console.print(f"[purple]({idx + 1}) {option.name}[/purple]")
        file_id_idx = typer.prompt("Please enter the budget index", type=int)
        assert file_id_idx - 1 in range(len(options)), "Did not select one of the options, exiting."
        server.set_file(options[file_id_idx - 1])
    else:
        server.set_file(file_id)
    file_id = server._file.file_id

    if not encryption_password and server._file.encrypt_key_id:
        encryption_password = typer.prompt("Please enter the encryption password for the budget", hide_input=True)
        # test the file
        server.download_budget(encryption_password)
    else:
        encryption_password = None

    if not context:
        # take the default context name as the file name in lowercase
        default_context = server._file.name.lower().replace(" ", "-")
        context = typer.prompt("Name of the context for this budget", default=default_context)

    config.budgets[context] = BudgetConfig(
        url=url,
        password=password,
        encryption_password=encryption_password,
        file_id=file_id,
    )
    if not config.default_context:
        config.default_context = context
    config.save()
    console.print(f"[green]Initialized budget '{context}'[/green]")


@app.command()
def use_context(context: str = typer.Argument(..., help="Context for this budget context")):
    """Sets the default context for the CLI."""
    if context not in config.budgets:
        raise ValueError(f"Context '{context}' is not registered. Choose one from {list(config.budgets.keys())}")
    config.default_context = context
    config.save()


@app.command()
def remove_context(context: str = typer.Argument(..., help="Context to be removed")):
    """Removes a configured context from the configuration."""
    if context not in config.budgets:
        raise ValueError(f"Context '{context}' is not registered. Choose one from {list(config.budgets.keys())}")
    config.budgets.pop(context)
    config.default_context = list(config.budgets.keys())[0] if len(config.budgets) == 1 else ""
    config.save()


@app.command()
def version():
    """
    Shows the library and server version.
    """
    actual = config.actual()
    info = actual.info()
    if state.output == OutputType.table:
        console.print(f"Library Version: {__version__}")
        console.print(f"Server Version: {info.build.version}")
    else:
        console.print(JSON.from_data({"library_version": __version__, "server_version": info.build.version}))


@app.command()
def accounts():
    """
    Show all accounts.
    """
    # Mock data for demonstration purposes
    accounts_data = []
    with config.actual() as actual:
        accounts_raw_data = get_accounts(actual.session)
        for account in accounts_raw_data:
            accounts_data.append(
                {
                    "name": account.name,
                    "balance": float(account.balance),
                }
            )

    if state.output == OutputType.table:
        table = Table(title="Accounts")
        table.add_column("Account Name", justify="left", style="cyan", no_wrap=True)
        table.add_column("Balance", justify="right", style="green")

        for account in accounts_data:
            table.add_row(account["name"], f"{account['balance']:.2f}")

        console.print(table)
    else:
        console.print(JSON.from_data(accounts_data))


@app.command()
def transactions():
    """
    Show all transactions.
    """
    transactions_data = []
    with config.actual() as actual:
        transactions_raw_data = get_transactions(actual.session)
        for transaction in transactions_raw_data:
            transactions_data.append(
                {
                    "date": transaction.get_date().isoformat(),
                    "payee": transaction.payee.name if transaction.payee else None,
                    "notes": transaction.notes or "",
                    "category": (transaction.category.name if transaction.category else None),
                    "amount": round(float(transaction.get_amount()), 2),
                }
            )

    if state.output == OutputType.table:
        table = Table(title="Transactions")
        table.add_column("Date", justify="left", style="cyan", no_wrap=True)
        table.add_column("Payee", justify="left", style="magenta")
        table.add_column("Notes", justify="left", style="yellow")
        table.add_column("Category", justify="left", style="cyan")
        table.add_column("Amount", justify="right", style="green")

        for transaction in transactions_data:
            color = "green" if transaction["amount"] >= 0 else "red"
            table.add_row(
                transaction["date"],
                transaction["payee"],
                transaction["notes"],
                transaction["category"],
                f"[{color}]{transaction['amount']:.2f}[/]",
            )

        console.print(table)
    else:
        console.print(JSON.from_data(transactions_data))


@app.command()
def payees():
    """
    Show all payees.
    """
    payees_data = []
    with config.actual() as actual:
        payees_raw_data = get_payees(actual.session)
        for payee in payees_raw_data:
            payees_data.append({"name": payee.name, "balance": round(float(payee.balance), 2)})

    if state.output == OutputType.table:
        table = Table(title="Payees")
        table.add_column("Name", justify="left", style="cyan", no_wrap=True)
        table.add_column("Balance", justify="right")

        for payee in payees_data:
            color = "green" if payee["balance"] >= 0 else "red"
            table.add_row(
                payee["name"],
                f"[{color}]{payee['balance']:.2f}[/]",
            )
        console.print(table)
    else:
        console.print(JSON.from_data(payees_data))


@app.command()
def export(
    filename: Optional[pathlib.Path] = typer.Argument(
        default=None,
        help="Name of the file to export, in zip format. "
        "Leave it empty to export it to the current folder with default name.",
    ),
):
    """
    Generates an export from the budget (for CLI backups).
    """
    with config.actual() as actual:
        if filename is None:
            current_date = datetime.datetime.now().strftime("%Y-%m-%d-%H%M")
            budget_name = actual.get_metadata().get("budgetName", "My Finances")
            filename = pathlib.Path(f"{current_date}-{budget_name}.zip")
        actual.export_data(filename)
        actual_metadata = actual.get_metadata()
        budget_name = actual_metadata["budgetName"]
        budget_id = actual_metadata["id"]
    console.print(
        f"[green]Exported budget '{budget_name}' (budget id '{budget_id}') to [bold]'{filename}'[/bold].[/green]"
    )


@app.command()
def metadata():
    """Displays all metadata for the current budget."""
    with config.actual() as actual:
        actual_metadata = actual.get_metadata()
    if state.output == OutputType.table:
        table = Table(title="Metadata")
        table.add_column("Key", justify="left", style="cyan", no_wrap=True)
        table.add_column("Value", justify="left")
        for key, value in actual_metadata.items():
            table.add_row(key, str(value))
        console.print(table)
    else:
        console.print(JSON.from_data(actual_metadata))


if __name__ == "__main__":
    app()



import re
import uuid
import warnings
from typing import List


def js_migration_statements(js_file: str) -> List[str]:
    queries = []
    matches = re.finditer(r"db\.(execQuery|runQuery)", js_file)
    for match in matches:
        start_index, end_index = match.regs[0][1], match.regs[0][1]
        # we now loop and find the first occasion where all parenthesis closed
        parenthesis_count, can_return = 0, False
        for i in range(start_index, len(js_file)):
            if js_file[i] == "(":
                can_return = True
                parenthesis_count += 1
            elif js_file[i] == ")":
                parenthesis_count -= 1
            if parenthesis_count == 0 and can_return:
                end_index = i + 1
                break
        function_call = js_file[start_index:end_index]
        # extract the query
        next_tick = function_call.find("`")
        next_quote = function_call.find("'")
        string_character = "`" if next_tick > 0 and ((next_tick < next_quote) or next_quote < 0) else "'"
        search = re.search(rf"{string_character}(.*?){string_character}", function_call, re.DOTALL)
        if not search:
            continue
        query = search.group(1)
        # skip empty queries
        if not query:
            continue
        # skip select queries
        if query.lower().startswith("select"):
            continue
        # if there are unknowns in the query, skip
        if "?" in query:
            warnings.warn(
                f"Migration query from migrations cannot be executed due to custom code, it will be skipped. Query:\n\n"
                f"{query}\n"
            )
            continue
        # if there is an uuid generation, use it
        while "${uuidv4()}" in query:
            query = query.replace("${uuidv4()}", str(uuid.uuid4()), 1)
        if not query.endswith(";"):
            query = query + ";"
        queries.append(query)
    return queries



import os
from enum import Enum
from pathlib import Path
from typing import Dict, Optional

import pydantic
import yaml
from rich.console import Console

from actual import Actual

console = Console()


def default_config_path():
    return Path.home() / ".actualpy" / "config.yaml"


class OutputType(Enum):
    table = "table"
    json = "json"


class State(pydantic.BaseModel):
    output: OutputType = pydantic.Field("table", alias="defaultOutput", description="Default output for CLI.")


class BudgetConfig(pydantic.BaseModel):
    url: str = pydantic.Field(..., description="")
    password: str = pydantic.Field(..., description="")
    file_id: str = pydantic.Field(..., alias="fileId")
    encryption_password: Optional[str] = pydantic.Field(None, alias="encryptionPassword")

    model_config = pydantic.ConfigDict(populate_by_name=True)


class Config(pydantic.BaseModel):
    default_context: str = pydantic.Field("", alias="defaultContext", description="Default budget context for CLI.")
    budgets: Dict[str, BudgetConfig] = pydantic.Field(
        default_factory=dict, description="Dict of configured budgets on CLI."
    )

    def save(self):
        """Saves the current configuration to a file."""
        config_path = default_config_path()
        os.makedirs(config_path.parent, exist_ok=True)
        os.makedirs(config_path.parent / "cache", exist_ok=True)
        with open(config_path, "w") as file:
            yaml.dump(self.model_dump(by_alias=True), file)

    @classmethod
    def load(cls):
        """Load the configuration file. If it doesn't exist, create a basic config."""
        config_path = default_config_path()
        if not config_path.exists():
            console.print(f"[yellow]Config file not found at '{config_path}'! Creating a new one...[/yellow]")
            # Create a basic config with default values
            default_config = cls()
            default_config.save()
            return default_config
        else:
            with open(config_path, "r") as file:
                config = yaml.safe_load(file)
                return cls.model_validate(config)

    def actual(self) -> Actual:
        context = self.default_context
        budget_config = self.budgets.get(context)
        if not budget_config:
            raise ValueError(f"Could not find budget with context '{context}'")
        return Actual(
            budget_config.url,
            password=budget_config.password,
            file=budget_config.file_id,
            encryption_password=budget_config.encryption_password,
            data_dir=default_config_path().parent / "cache" / context,
        )



import zipfile
from unittest.mock import patch

import pytest
from requests import Session

from actual import Actual, reflect_model
from actual.api import ListUserFilesDTO
from actual.api.models import RemoteFileListDTO, StatusCode
from actual.exceptions import ActualError, AuthorizationError, UnknownFileId
from actual.protobuf_models import Message
from tests.conftest import RequestsMock


def test_api_apply(mocker, session):
    mocker.patch("actual.Actual.validate")
    actual = Actual(token="foo")
    actual.engine = session.bind
    actual._meta = reflect_model(session.bind)
    # not found table
    m = Message(dict(dataset="foo", row="foobar", column="bar"))
    m.set_value("foobar")
    with pytest.raises(ActualError, match="table 'foo' not found"):
        actual.apply_changes([m])
    m.dataset = "accounts"
    with pytest.raises(ActualError, match="column 'bar' at table 'accounts' not found"):
        actual.apply_changes([m])


def test_rename_delete_budget_without_file(mocker):
    mocker.patch("actual.Actual.validate")
    actual = Actual(token="foo")
    actual._file = None
    with pytest.raises(UnknownFileId, match="No current file loaded"):
        actual.delete_budget()
    with pytest.raises(UnknownFileId, match="No current file loaded"):
        actual.rename_budget("foo")


@patch.object(Session, "post", return_value=RequestsMock({"status": "error", "reason": "proxy-not-trusted"}))
def test_api_login_unknown_error(_post, mocker):
    mocker.patch("actual.Actual.validate")
    actual = Actual(token="foo")
    actual.api_url = "localhost"
    actual.cert = False
    with pytest.raises(AuthorizationError, match="Something went wrong on login"):
        actual.login("foo")


@patch.object(Session, "post", return_value=RequestsMock({}, status_code=403))
def test_api_login_http_error(_post, mocker):
    mocker.patch("actual.Actual.validate")
    actual = Actual(token="foo")
    actual.api_url = "localhost"
    actual.cert = False
    with pytest.raises(AuthorizationError, match="HTTP error '403'"):
        actual.login("foo")


def test_no_certificate(mocker):
    mocker.patch("actual.Actual.validate")
    actual = Actual(token="foo", cert=False)
    assert actual._requests_session.verify is False


def test_set_file_exceptions(mocker):
    mocker.patch("actual.Actual.validate")
    list_user_files = mocker.patch(
        "actual.Actual.list_user_files", return_value=ListUserFilesDTO(status=StatusCode.OK, data=[])
    )
    actual = Actual(token="foo")
    with pytest.raises(ActualError, match="Could not find a file id or identifier 'foo'"):
        actual.set_file("foo")
    list_user_files.return_value = ListUserFilesDTO(
        status=StatusCode.OK,
        data=[
            RemoteFileListDTO(deleted=False, fileId="foo", groupId="foo", name="foo", encryptKeyId=None),
            RemoteFileListDTO(deleted=False, fileId="foo", groupId="foo", name="foo", encryptKeyId=None),
        ],
    )
    with pytest.raises(ActualError, match="Multiple files found with identifier 'foo'"):
        actual.set_file("foo")


def test_zip_exceptions(mocker, tmp_path):
    mocker.patch("actual.Actual.validate")
    mocker.patch("actual.Actual.create_engine")
    archive = tmp_path / "file.zip"
    with zipfile.ZipFile(archive, "w"):
        pass
    actual = Actual(token="foo")
    actual.import_zip(archive)
    # archive will use a normal temp folder since the cloudFileId is missing from metadata
    assert actual._data_dir.name.startswith("tmp")


def test_api_extra_headers(mocker):
    mocker.patch("actual.Actual.validate")
    actual = Actual(token="foo", extra_headers={"foo": "bar"})
    assert actual._requests_session.headers["foo"] == "bar"
    assert actual._requests_session.headers["X-ACTUAL-TOKEN"] == "foo"



from __future__ import annotations

import base64
import datetime
import uuid
from typing import List

import proto

from actual.crypto import decrypt, encrypt
from actual.exceptions import ActualDecryptionError

"""
Protobuf message definitions taken from the [sync.proto file](
https://github.com/actualbudget/actual/blob/029e2f09bf6caf386523bbfa944ab845271a3932/packages/crdt/src/proto/sync.proto).

They should represent how the server take requests from the client. The server side implementation is available [here](
https://github.com/actualbudget/actual-server/blob/master/src/app-sync.js#L32).
"""


class HULC_Client:
    def __init__(self, client_id: str = None, initial_count: int = 0, ts: datetime.datetime = None):
        self.client_id = client_id or self.random_client_id()
        self.initial_count = initial_count
        self.ts = ts or datetime.datetime(1970, 1, 1, 0, 0, 0)

    @classmethod
    def from_timestamp(cls, ts: str) -> HULC_Client:
        ts_string, _, rest = ts.partition("Z")
        segments = rest.split("-")
        ts = datetime.datetime.fromisoformat(ts_string)
        return cls(segments[-1], int(segments[-2], 16), ts)

    def __str__(self):
        count = f"{self.initial_count:0>4X}"
        return f"{self.ts.isoformat(timespec='milliseconds')}Z-{count}-{self.client_id}"

    def timestamp(self, now: datetime.datetime = None) -> str:
        """Actual uses Hybrid Unique Logical Clock (HULC) timestamp generator.

        Timestamps serialize into a 46-character collatable string. Examples:

         - `2015-04-24T22:23:42.123Z-1000-0123456789ABCDEF`
         - `2015-04-24T22:23:42.123Z-1000-A219E7A71CC18912`

        See [original source code](
        https://github.com/actualbudget/actual/blob/a9362cc6f9b974140a760ad05816cac51c849769/packages/crdt/src/crdt/timestamp.ts)
        for reference.
        """
        if not now:
            now = datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None)
        count = f"{self.initial_count:0>4X}"
        self.initial_count += 1
        return f"{now.isoformat(timespec='milliseconds')}Z-{count}-{self.client_id}"

    @staticmethod
    def random_client_id():
        """Creates a client id for the HULC request. Implementation copied [from the source code](
        https://github.com/actualbudget/actual/blob/a9362cc6f9b974140a760ad05816cac51c849769/packages/crdt/src/crdt/timestamp.ts#L80)
        """
        return str(uuid.uuid4()).replace("-", "")[-16:]


class EncryptedData(proto.Message):
    iv = proto.Field(proto.BYTES, number=1)
    authTag = proto.Field(proto.BYTES, number=2)
    data = proto.Field(proto.BYTES, number=3)


class Message(proto.Message):
    dataset = proto.Field(proto.STRING, number=1)
    row = proto.Field(proto.STRING, number=2)
    column = proto.Field(proto.STRING, number=3)
    value = proto.Field(proto.STRING, number=4)

    def get_value(self) -> str | int | float | None:
        """Serialization types from Actual. [Original source code](
        https://github.com/actualbudget/actual/blob/998efb9447da6f8ce97956cbe83d6e8a3c18cf53/packages/loot-core/src/server/sync/index.ts#L154-L160)
        """
        datatype, _, value = self.value.partition(":")
        if datatype == "S":
            return value
        elif datatype == "N":
            return float(value)
        elif datatype == "0":
            return None
        else:
            raise ValueError(f"Conversion not supported for datatype '{datatype}'")

    def set_value(self, value: str | int | float | None) -> str:
        if isinstance(value, str):
            datatype = "S"
        elif isinstance(value, int) or isinstance(value, float):
            datatype = "N"
        elif value is None:
            datatype = "0"
        else:
            raise ValueError(f"Conversion not supported for datatype '{type(value)}'")
        self.value = f"{datatype}:{value}"
        return self.value


class MessageEnvelope(proto.Message):
    timestamp = proto.Field(proto.STRING, number=1)
    isEncrypted = proto.Field(proto.BOOL, number=2)
    content = proto.Field(proto.BYTES, number=3)

    def set_timestamp(self, client_id: str = None, now: datetime.datetime = None) -> str:
        self.timestamp = HULC_Client(client_id).timestamp(now)
        return self.timestamp


class SyncRequest(proto.Message):
    messages = proto.RepeatedField(MessageEnvelope, number=1)
    fileId = proto.Field(proto.STRING, number=2)
    groupId = proto.Field(proto.STRING, number=3)
    keyId = proto.Field(proto.STRING, number=5)
    since = proto.Field(proto.STRING, number=6)

    def set_timestamp(self, client_id: str = None, now: datetime.datetime = None) -> str:
        self.since = HULC_Client(client_id).timestamp(now)
        return self.since

    def set_null_timestamp(self, client_id: str = None) -> str:
        return self.set_timestamp(client_id, datetime.datetime(1970, 1, 1, 0, 0, 0, 0))

    def set_messages(self, messages: List[Message], client: HULC_Client, master_key: bytes = None):
        if not self.messages:
            self.messages = []
        for message in messages:
            content = Message.serialize(message)
            is_encrypted = False
            if master_key is not None:
                encrypted_content = encrypt("", master_key, content)
                encrypted_data = EncryptedData(
                    {
                        "iv": base64.b64decode(encrypted_content["meta"]["iv"]),
                        "authTag": base64.b64decode(encrypted_content["meta"]["authTag"]),
                        "data": base64.b64decode(encrypted_content["value"]),
                    }
                )
                content = EncryptedData.serialize(encrypted_data)
                is_encrypted = True
            m = MessageEnvelope({"content": content, "isEncrypted": is_encrypted})
            m.timestamp = client.timestamp()
            self.messages.append(m)


class SyncResponse(proto.Message):
    messages = proto.RepeatedField(MessageEnvelope, number=1)
    merkle = proto.Field(proto.STRING, number=2)

    def get_messages(self, master_key: bytes = None) -> List[Message]:
        messages = []
        for message in self.messages:  # noqa
            if message.isEncrypted:
                if not master_key:
                    raise ActualDecryptionError("Master key not provided and data is encrypted.")
                encrypted = EncryptedData.deserialize(message.content)
                content = decrypt(master_key, encrypted.iv, encrypted.data, encrypted.authTag)
            else:
                content = message.content
            messages.append(Message.deserialize(content))
        return messages



"""
This file was partially generated using sqlacodegen using the downloaded version of the db.sqlite file export
in order to update this file, you can generate the code with:

```bash
sqlacodegen --generator sqlmodels sqlite:///db.sqlite
```

and patch the necessary models by merging the results. The [BaseModel][actual.database.BaseModel] defines all models
that can be updated from the user, and must contain a unique `id`. Those models can then be converted automatically
into a protobuf change message using [BaseModel.convert][actual.database.BaseModel.convert].

It is preferred to create database entries using the [queries][actual.queries], rather than using the raw database
model.
"""

import datetime
import decimal
import json
from typing import Dict, List, Optional, Tuple, Union

from sqlalchemy import MetaData, Table, engine, event, inspect
from sqlalchemy.dialects.sqlite import insert
from sqlalchemy.orm import class_mapper, object_session
from sqlmodel import (
    Boolean,
    Column,
    Field,
    Float,
    ForeignKey,
    Index,
    Integer,
    LargeBinary,
    Relationship,
    Session,
    SQLModel,
    Text,
    func,
    select,
    text,
)

from actual.exceptions import ActualInvalidOperationError
from actual.protobuf_models import HULC_Client, Message
from actual.utils.conversions import cents_to_decimal, date_to_int, decimal_to_cents, int_to_date, month_range

"""
This variable contains the internal model mappings for all databases. It solves a couple of issues, namely having the
mapping from `__tablename__` to the actual SQLAlchemy class, and later mapping the SQL column into the Pydantic field,
which could be different and follows the Python naming convention. An example is the field `Transactions.is_parent`,
that converts into the SQL equivalent `transactions.isParent`. In this case, we would have the following entries:

```
__TABLE_COLUMNS_MAP__ = {
    "transactions": {
        "entity": <class 'actual.database.Transactions'>,
        "columns": {
            "isParent": "is_parent"
        }
    }
}
```
"""
__TABLE_COLUMNS_MAP__ = dict()


def reflect_model(eng: engine.Engine) -> MetaData:
    """Reflects the current state of the database."""
    local_meta = MetaData()
    local_meta.reflect(bind=eng)
    return local_meta


def get_class_from_reflected_table_name(metadata: MetaData, table_name: str) -> Union[Table, None]:
    """
    Returns, based on the defined tables on the reflected model the corresponding SQLAlchemy table.
    If not found, returns `None`.
    """
    return metadata.tables.get(table_name, None)


def get_attribute_from_reflected_table_name(
    metadata: MetaData, table_name: str, column_name: str
) -> Union[Column, None]:
    """
    Returns, based, on the defined reflected model the corresponding and the SAColumn. If not found, returns `None`.
    """
    table = get_class_from_reflected_table_name(metadata, table_name)
    return table.columns.get(column_name, None)


def get_class_by_table_name(table_name: str) -> Union[SQLModel, None]:
    """
    Returns, based on the defined tables `__tablename__` the corresponding SQLModel object. If not found, returns
    `None`.
    """
    return __TABLE_COLUMNS_MAP__.get(table_name, {}).get("entity", None)


def get_attribute_by_table_name(table_name: str, column_name: str, reverse: bool = False) -> Union[str, None]:
    """
    Returns, based, on the defined tables `__tablename__` and the SAColumn name, the correct pydantic attribute. Search
    can be reversed by setting the `reverse` flag to `True`.
    If not found, returns `None`.

    :param table_name: SQL table name.
    :param column_name: SQL column name.
    :param reverse: If true, reverses the search and returns the SAColumn from the Pydantic attribute.
    :return: Pydantic attribute name or SAColumn name.
    """
    return (
        __TABLE_COLUMNS_MAP__.get(table_name, {})
        .get("columns" if not reverse else "rev_columns", {})
        .get(column_name, None)
    )


def apply_change(
    session: Session, table: Table, table_id: str, values: Dict[Column, Union[str, int, float, None]]
) -> None:
    """This function upserts multiple changes into a table based on the `table_id` as primary key. All the `values`
    will be inserted as a new row, and if the id already exists, the values will be updated."""
    insert_stmt = (
        insert(table).values({"id": table_id, **values}).on_conflict_do_update(index_elements=["id"], set_=values)
    )
    session.exec(insert_stmt)  # noqa: Insert type here is correct


def strong_reference_session(session: Session):
    @event.listens_for(session, "before_flush")
    def before_flush(sess, flush_context, instances):
        if len(sess.deleted):
            raise ActualInvalidOperationError(
                "Actual does not allow deleting entries, set the `tombstone` to 1 instead or call the .delete() method"
            )
        if "messages" not in sess.info:
            sess.info["messages"] = messages = []
        else:
            messages = sess.info["messages"]
        # convert entries from the model
        for instance in sess.new:
            # all entries that were added new
            messages.extend(instance.convert(is_new=True))
        for instance in sess.dirty:
            # all entries that were modified
            messages.extend(instance.convert(is_new=False))

    @event.listens_for(session, "after_commit")
    @event.listens_for(session, "after_soft_rollback")
    def after_commit_or_rollback(
        sess,
        previous_transaction=None,  # noqa: previous_transaction needed for soft rollback
    ):
        if sess.info.get("messages"):
            del sess.info["messages"]

    return session


class BaseModel(SQLModel):
    id: str = Field(sa_column=Column("id", Text, primary_key=True))

    def convert(self, is_new: bool = True) -> List[Message]:
        """Convert the object into distinct entries for sync method. Based on the [original implementation](
        https://github.com/actualbudget/actual/blob/98c17bd5e0f13e27a09a7f6ac176510530572be7/packages/loot-core/src/server/aql/schema-helpers.ts#L146)
        """
        row = getattr(self, "id", None)  # also helps lazy loading the instance
        if row is None:
            raise AttributeError(
                f"Cannot convert model {self.__name__} because it misses the 'id' attribute.\n"
                f"If you see this error, make sure your entry has a unique 'id' as primary key."
            )
        # compute changes from a sqlalchemy instance, see https://stackoverflow.com/a/28353846/12681470
        changes = []
        for column in self.changed():
            converted_attr_name = get_attribute_by_table_name(self.__tablename__, column, reverse=True)
            m = Message(dict(dataset=self.__tablename__, row=row, column=converted_attr_name))
            value = self.__getattribute__(column)
            # if the entry is new, we can ignore null columns, otherwise consider it an update to None
            if value is not None or not is_new:
                m.set_value(value)
                changes.append(m)
        return changes

    def changed(self) -> List[str]:
        """Returns list of model changed attributes."""
        changed_attributes = []
        inspr = inspect(self)
        attrs = class_mapper(self.__class__).column_attrs  # exclude relationships
        for attr in attrs:  # noqa: you can iterate over attrs
            column = attr.key
            if column == "id":
                continue
            hist = getattr(inspr.attrs, column).history
            if hist.has_changes():
                changed_attributes.append(column)
        return changed_attributes

    def delete(self):
        """Deletes the model, by setting the `tombstone` attribute to 1. It is only possible to hard delete
        transactions by updating and re-uploading the downloaded budget."""
        if not hasattr(self, "tombstone"):
            raise AttributeError(f"Model {self.__class__.__name__} has no tombstone field and cannot be deleted.")
        setattr(self, "tombstone", 1)


class Meta(SQLModel, table=True):
    __tablename__ = "__meta__"

    key: Optional[str] = Field(default=None, sa_column=Column("key", Text, primary_key=True))
    value: Optional[str] = Field(default=None, sa_column=Column("value", Text))


class Migrations(SQLModel, table=True):
    __tablename__ = "__migrations__"

    id: Optional[int] = Field(default=None, sa_column=Column("id", Integer, primary_key=True))


class Accounts(BaseModel, table=True):
    id: Optional[str] = Field(default=None, sa_column=Column("id", Text, primary_key=True))
    account_id: Optional[str] = Field(default=None, sa_column=Column("account_id", Text))
    name: Optional[str] = Field(default=None, sa_column=Column("name", Text))
    # Careful when using those balance fields, are they might be empty. Use account.balance property instead
    balance_current: Optional[int] = Field(default=None, sa_column=Column("balance_current", Integer))
    balance_available: Optional[int] = Field(default=None, sa_column=Column("balance_available", Integer))
    balance_limit: Optional[int] = Field(default=None, sa_column=Column("balance_limit", Integer))
    mask: Optional[str] = Field(default=None, sa_column=Column("mask", Text))
    official_name: Optional[str] = Field(default=None, sa_column=Column("official_name", Text))
    subtype: Optional[str] = Field(default=None, sa_column=Column("subtype", Text))
    bank_id: Optional[str] = Field(default=None, sa_column=Column("bank", Text, ForeignKey("banks.id")))
    offbudget: Optional[int] = Field(default=None, sa_column=Column("offbudget", Integer, server_default=text("0")))
    closed: Optional[int] = Field(default=None, sa_column=Column("closed", Integer, server_default=text("0")))
    tombstone: Optional[int] = Field(default=None, sa_column=Column("tombstone", Integer, server_default=text("0")))
    sort_order: Optional[float] = Field(default=None, sa_column=Column("sort_order", Float))
    type: Optional[str] = Field(default=None, sa_column=Column("type", Text))
    account_sync_source: Optional[str] = Field(default=None, sa_column=Column("account_sync_source", Text))
    last_sync: Optional[str] = Field(default=None, sa_column=Column("last_sync", Text))
    last_reconciled: Optional[str] = Field(default=None, sa_column=Column("last_reconciled", Text))

    payee: "Payees" = Relationship(back_populates="account", sa_relationship_kwargs={"uselist": False})
    transactions: List["Transactions"] = Relationship(
        back_populates="account",
        sa_relationship_kwargs={
            "primaryjoin": (
                "and_(Accounts.id == Transactions.acct,Transactions.is_parent == 0, Transactions.tombstone==0)"
            )
        },
    )
    bank: "Banks" = Relationship(
        back_populates="account",
        sa_relationship_kwargs={
            "uselist": False,
            "primaryjoin": "and_(Accounts.bank_id == Banks.id,Banks.tombstone == 0)",
        },
    )

    @property
    def balance(self) -> decimal.Decimal:
        """Returns the current balance of the account. Deleted transactions are ignored."""
        value = object_session(self).scalar(
            select(func.coalesce(func.sum(Transactions.amount), 0)).where(
                Transactions.acct == self.id,
                Transactions.is_parent == 0,
                Transactions.tombstone == 0,
            )
        )
        return cents_to_decimal(value)

    @property
    def notes(self) -> Optional[str]:
        """Returns notes for the account. If none are present, returns `None`."""
        return object_session(self).scalar(select(Notes.note).where(Notes.id == f"account-{self.id}"))


class Banks(BaseModel, table=True):
    id: Optional[str] = Field(default=None, sa_column=Column("id", Text, primary_key=True))
    bank_id: Optional[str] = Field(default=None, sa_column=Column("bank_id", Text))
    name: Optional[str] = Field(default=None, sa_column=Column("name", Text))
    tombstone: Optional[int] = Field(default=None, sa_column=Column("tombstone", Integer, server_default=text("0")))

    account: "Accounts" = Relationship(back_populates="bank")


class Categories(BaseModel, table=True):
    """
    Stores the category list, which is the classification applied on top of the transaction.

    Each category will belong to its own category group.
    """

    hidden: bool = Field(sa_column=Column("hidden", Boolean, nullable=False, server_default=text("0")))
    id: Optional[str] = Field(default=None, sa_column=Column("id", Text, primary_key=True))
    name: Optional[str] = Field(default=None, sa_column=Column("name", Text))
    is_income: Optional[int] = Field(default=None, sa_column=Column("is_income", Integer, server_default=text("0")))
    cat_group: Optional[str] = Field(
        default=None, sa_column=Column("cat_group", Text, ForeignKey("category_groups.id"))
    )
    sort_order: Optional[float] = Field(default=None, sa_column=Column("sort_order", Float))
    tombstone: Optional[int] = Field(default=None, sa_column=Column("tombstone", Integer, server_default=text("0")))
    goal_def: Optional[str] = Field(default=None, sa_column=Column("goal_def", Text, server_default=text("null")))

    zero_budgets: "ZeroBudgets" = Relationship(
        back_populates="category",
        sa_relationship_kwargs={
            "primaryjoin": "and_(ZeroBudgets.category_id == Categories.id)",
        },
    )
    reflect_budgets: "ReflectBudgets" = Relationship(
        back_populates="category",
        sa_relationship_kwargs={
            "primaryjoin": "and_(ReflectBudgets.category_id == Categories.id)",
        },
    )
    transactions: List["Transactions"] = Relationship(
        back_populates="category",
        sa_relationship_kwargs={
            "primaryjoin": (
                "and_(Categories.id == Transactions.category_id,Transactions.is_parent == 0, Transactions.tombstone==0)"
            )
        },
    )
    group: "CategoryGroups" = Relationship(
        back_populates="categories",
        sa_relationship_kwargs={
            "primaryjoin": "and_(Categories.cat_group == CategoryGroups.id, CategoryGroups.tombstone == 0)",
            "uselist": False,
        },
    )

    @property
    def balance(self) -> decimal.Decimal:
        """Returns the current balance of the category. Deleted transactions are ignored."""
        value = object_session(self).scalar(
            select(func.coalesce(func.sum(Transactions.amount), 0)).where(
                Transactions.category_id == self.id,
                Transactions.is_parent == 0,
                Transactions.tombstone == 0,
            )
        )
        return cents_to_decimal(value)


class CategoryGroups(BaseModel, table=True):
    """
    Stores the groups that the categories can belong to.
    """

    __tablename__ = "category_groups"

    hidden: bool = Field(sa_column=Column("hidden", Boolean, nullable=False, server_default=text("0")))
    id: Optional[str] = Field(default=None, sa_column=Column("id", Text, primary_key=True))
    name: Optional[str] = Field(default=None, sa_column=Column("name", Text))
    is_income: Optional[int] = Field(default=None, sa_column=Column("is_income", Integer, server_default=text("0")))
    sort_order: Optional[float] = Field(default=None, sa_column=Column("sort_order", Float))
    tombstone: Optional[int] = Field(default=None, sa_column=Column("tombstone", Integer, server_default=text("0")))

    categories: List["Categories"] = Relationship(
        back_populates="group",
        sa_relationship_kwargs={
            "primaryjoin": "and_(CategoryGroups.id == Categories.cat_group, Categories.tombstone == 0)",
        },
    )


class CategoryMapping(BaseModel, table=True):
    __tablename__ = "category_mapping"

    id: Optional[str] = Field(default=None, sa_column=Column("id", Text, primary_key=True))
    transfer_id: Optional[str] = Field(default=None, sa_column=Column("transferId", Text))


class CreatedBudgets(SQLModel, table=True):
    __tablename__ = "created_budgets"

    month: Optional[str] = Field(default=None, sa_column=Column("month", Text, primary_key=True))


class CustomReports(BaseModel, table=True):
    __tablename__ = "custom_reports"

    id: Optional[str] = Field(default=None, sa_column=Column("id", Text, primary_key=True))
    name: Optional[str] = Field(default=None, sa_column=Column("name", Text))
    start_date: Optional[str] = Field(default=None, sa_column=Column("start_date", Text))
    end_date: Optional[str] = Field(default=None, sa_column=Column("end_date", Text))
    date_static: Optional[int] = Field(default=None, sa_column=Column("date_static", Integer, server_default=text("0")))
    date_range: Optional[str] = Field(default=None, sa_column=Column("date_range", Text))
    mode: Optional[str] = Field(default=None, sa_column=Column("mode", Text, server_default=text("'total'")))
    group_by: Optional[str] = Field(default=None, sa_column=Column("group_by", Text, server_default=text("'Category'")))
    balance_type: Optional[str] = Field(
        default=None, sa_column=Column("balance_type", Text, server_default=text("'Expense'"))
    )
    show_empty: Optional[int] = Field(default=None, sa_column=Column("show_empty", Integer, server_default=text("0")))
    show_offbudget: Optional[int] = Field(
        default=None, sa_column=Column("show_offbudget", Integer, server_default=text("0"))
    )
    show_hidden: Optional[int] = Field(default=None, sa_column=Column("show_hidden", Integer, server_default=text("0")))
    show_uncategorized: Optional[int] = Field(
        default=None, sa_column=Column("show_uncategorized", Integer, server_default=text("0"))
    )
    selected_categories: Optional[str] = Field(default=None, sa_column=Column("selected_categories", Text))
    graph_type: Optional[str] = Field(
        default=None, sa_column=Column("graph_type", Text, server_default=text("'BarGraph'"))
    )
    conditions: Optional[str] = Field(default=None, sa_column=Column("conditions", Text))
    conditions_op: Optional[str] = Field(
        default=None, sa_column=Column("conditions_op", Text, server_default=text("'and'"))
    )
    metadata_: Optional[str] = Field(default=None, sa_column=Column("metadata", Text))
    interval: Optional[str] = Field(default=None, sa_column=Column("interval", Text, server_default=text("'Monthly'")))
    color_scheme: Optional[str] = Field(default=None, sa_column=Column("color_scheme", Text))
    tombstone: Optional[int] = Field(default=None, sa_column=Column("tombstone", Integer, server_default=text("0")))
    include_current: Optional[int] = Field(
        default=None, sa_column=Column("include_current", Integer, server_default=text("0"))
    )
    sort_by: Optional[str] = Field(default=None, sa_column=Column("sort_by", Text, server_default=text("'desc'")))


class Dashboard(BaseModel, table=True):
    id: Optional[str] = Field(default=None, sa_column=Column("id", Text, primary_key=True))
    type: Optional[str] = Field(default=None, sa_column=Column("type", Text))
    width: Optional[int] = Field(default=None, sa_column=Column("width", Integer))
    height: Optional[int] = Field(default=None, sa_column=Column("height", Integer))
    x: Optional[int] = Field(default=None, sa_column=Column("x", Integer))
    y: Optional[int] = Field(default=None, sa_column=Column("y", Integer))
    meta: Optional[str] = Field(default=None, sa_column=Column("meta", Text))
    tombstone: Optional[int] = Field(default=None, sa_column=Column("tombstone", Integer, server_default=text("0")))


class Kvcache(SQLModel, table=True):
    key: Optional[str] = Field(default=None, sa_column=Column("key", Text, primary_key=True))
    value: Optional[str] = Field(default=None, sa_column=Column("value", Text))


class KvcacheKey(SQLModel, table=True):
    __tablename__ = "kvcache_key"

    id: Optional[int] = Field(default=None, sa_column=Column("id", Integer, primary_key=True))
    key: Optional[float] = Field(default=None, sa_column=Column("key", Float))


class MessagesClock(SQLModel, table=True):
    __tablename__ = "messages_clock"

    id: Optional[int] = Field(default=None, sa_column=Column("id", Integer, primary_key=True))
    clock: Optional[str] = Field(default=None, sa_column=Column("clock", Text))

    def get_clock(self) -> dict:
        """Gets the clock from JSON text to a dictionary with fields `timestamp` and `merkle`."""
        return json.loads(self.clock)

    def set_clock(self, value: dict):
        """Sets the clock from a dictionary and stores it in the correct format."""
        self.clock = json.dumps(value, separators=(",", ":"))

    def get_timestamp(self) -> HULC_Client:
        """Gets the timestamp from the clock value directly as a [HULC_Client][actual.protobuf_models.HULC_Client]."""
        clock = self.get_clock()
        return HULC_Client.from_timestamp(clock["timestamp"])

    def set_timestamp(self, client: HULC_Client) -> None:
        """Sets the timestamp on the clock value based on the [HULC_Client][actual.protobuf_models.HULC_Client]
        provided."""
        clock_message = self.get_clock()
        clock_message["timestamp"] = str(client)
        self.set_clock(clock_message)


class MessagesCrdt(SQLModel, table=True):
    __tablename__ = "messages_crdt"
    __table_args__ = (Index("messages_crdt_search", "dataset", "row", "column", "timestamp"),)

    timestamp: str = Field(sa_column=Column("timestamp", Text, nullable=False, unique=True))
    dataset: str = Field(sa_column=Column("dataset", Text, nullable=False))
    row: str = Field(sa_column=Column("row", Text, nullable=False))
    column: str = Field(sa_column=Column("column", Text, nullable=False))
    value: bytes = Field(sa_column=Column("value", LargeBinary, nullable=False))
    id: Optional[int] = Field(default=None, sa_column=Column("id", Integer, primary_key=True))


class Notes(BaseModel, table=True):
    """Stores the description of each account."""

    id: Optional[str] = Field(default=None, sa_column=Column("id", Text, primary_key=True))
    note: Optional[str] = Field(default=None, sa_column=Column("note", Text))


class PayeeMapping(BaseModel, table=True):
    __tablename__ = "payee_mapping"

    id: Optional[str] = Field(default=None, sa_column=Column("id", Text, primary_key=True))
    target_id: Optional[str] = Field(default=None, sa_column=Column("targetId", Text))


class Payees(BaseModel, table=True):
    """
    Stores the individual payees.

    Each payee is a unique identifier that can be assigned to a transaction. Certain payees have empty names and are
    associated to the accounts themselves, representing the transfer between one account and another. These would
    have the field `account` not set to `None`.
    """

    id: Optional[str] = Field(default=None, sa_column=Column("id", Text, primary_key=True))
    name: Optional[str] = Field(default=None, sa_column=Column("name", Text))
    category: Optional[str] = Field(default=None, sa_column=Column("category", Text))
    tombstone: Optional[int] = Field(default=None, sa_column=Column("tombstone", Integer, server_default=text("0")))
    transfer_acct: Optional[str] = Field(
        default=None, sa_column=Column("transfer_acct", Text, ForeignKey("accounts.id"))
    )
    favorite: Optional[int] = Field(default=None, sa_column=Column("favorite", Integer, server_default=text("0")))
    learn_categories: Optional[bool] = Field(sa_column=Column("learn_categories", Boolean, server_default=text("1")))

    account: Optional["Accounts"] = Relationship(back_populates="payee", sa_relationship_kwargs={"uselist": False})
    transactions: List["Transactions"] = Relationship(
        back_populates="payee",
        sa_relationship_kwargs={"primaryjoin": "and_(Transactions.payee_id == Payees.id, Transactions.tombstone==0)"},
    )

    @property
    def balance(self) -> decimal.Decimal:
        """Returns the current balance of the payee. Deleted transactions are ignored."""
        value = object_session(self).scalar(
            select(func.coalesce(func.sum(Transactions.amount), 0)).where(
                Transactions.payee_id == self.id,
                Transactions.is_parent == 0,
                Transactions.tombstone == 0,
            )
        )
        return cents_to_decimal(value)


class Preferences(BaseModel, table=True):
    """Stores the preferences for the user, using key/value pairs."""

    id: Optional[str] = Field(default=None, sa_column=Column("id", Text, primary_key=True))
    value: Optional[str] = Field(default=None, sa_column=Column("value", Text))


class Rules(BaseModel, table=True):
    """
    Stores all rules on the budget. The conditions and actions are stored separately using the JSON format.

    The conditions are stored as a text field, but can be retrieved as a model using
    [get_ruleset][actual.queries.get_ruleset].
    """

    id: Optional[str] = Field(default=None, sa_column=Column("id", Text, primary_key=True))
    stage: Optional[str] = Field(default=None, sa_column=Column("stage", Text))
    conditions: Optional[str] = Field(default=None, sa_column=Column("conditions", Text))
    actions: Optional[str] = Field(default=None, sa_column=Column("actions", Text))
    tombstone: Optional[int] = Field(default=None, sa_column=Column("tombstone", Integer, server_default=text("0")))
    conditions_op: Optional[str] = Field(
        default=None,
        sa_column=Column("conditions_op", Text, server_default=text("'and'")),
    )


class Schedules(SQLModel, table=True):
    """Stores the schedules defined by the user. Is also linked to a rule that executes it."""

    id: Optional[str] = Field(default=None, sa_column=Column("id", Text, primary_key=True))
    rule_id: Optional[str] = Field(default=None, sa_column=Column("rule", Text, ForeignKey("rules.id")))
    active: Optional[int] = Field(default=None, sa_column=Column("active", Integer, server_default=text("0")))
    completed: Optional[int] = Field(default=None, sa_column=Column("completed", Integer, server_default=text("0")))
    posts_transaction: Optional[int] = Field(
        default=None,
        sa_column=Column("posts_transaction", Integer, server_default=text("0")),
    )
    tombstone: Optional[int] = Field(default=None, sa_column=Column("tombstone", Integer, server_default=text("0")))
    name: Optional[str] = Field(default=None, sa_column=Column("name", Text, server_default=text("NULL")))

    rule: "Rules" = Relationship(sa_relationship_kwargs={"uselist": False})
    transactions: List["Transactions"] = Relationship(back_populates="schedule")


class SchedulesJsonPaths(SQLModel, table=True):
    __tablename__ = "schedules_json_paths"

    schedule_id: Optional[str] = Field(default=None, sa_column=Column("schedule_id", Text, primary_key=True))
    payee: Optional[str] = Field(default=None, sa_column=Column("payee", Text))
    account: Optional[str] = Field(default=None, sa_column=Column("account", Text))
    amount: Optional[str] = Field(default=None, sa_column=Column("amount", Text))
    date: Optional[str] = Field(default=None, sa_column=Column("date", Text))


class SchedulesNextDate(SQLModel, table=True):
    __tablename__ = "schedules_next_date"

    id: Optional[str] = Field(default=None, sa_column=Column("id", Text, primary_key=True))
    schedule_id: Optional[str] = Field(default=None, sa_column=Column("schedule_id", Text))
    local_next_date: Optional[int] = Field(default=None, sa_column=Column("local_next_date", Integer))
    local_next_date_ts: Optional[int] = Field(default=None, sa_column=Column("local_next_date_ts", Integer))
    base_next_date: Optional[int] = Field(default=None, sa_column=Column("base_next_date", Integer))
    base_next_date_ts: Optional[int] = Field(default=None, sa_column=Column("base_next_date_ts", Integer))
    tombstone: Optional[int] = Field(default=None, sa_column=Column("tombstone", Integer, server_default=text("0")))


class TransactionFilters(BaseModel, table=True):
    __tablename__ = "transaction_filters"

    id: Optional[str] = Field(default=None, sa_column=Column("id", Text, primary_key=True))
    name: Optional[str] = Field(default=None, sa_column=Column("name", Text))
    conditions: Optional[str] = Field(default=None, sa_column=Column("conditions", Text))
    conditions_op: Optional[str] = Field(
        default=None,
        sa_column=Column("conditions_op", Text, server_default=text("'and'")),
    )
    tombstone: Optional[int] = Field(default=None, sa_column=Column("tombstone", Integer, server_default=text("0")))


class Transactions(BaseModel, table=True):
    """
    Contains all transactions inserted into Actual.
    """

    __table_args__ = (
        Index("trans_category", "category"),
        Index("trans_category_date", "category", "date"),
        Index("trans_date", "date"),
        Index("trans_parent_id", "parent_id"),
        Index("trans_sorted", "date", "starting_balance_flag", "sort_order", "id"),
    )

    id: Optional[str] = Field(default=None, sa_column=Column("id", Text, primary_key=True))
    is_parent: Optional[int] = Field(default=None, sa_column=Column("isParent", Integer, server_default=text("0")))
    is_child: Optional[int] = Field(default=None, sa_column=Column("isChild", Integer, server_default=text("0")))
    acct: Optional[str] = Field(default=None, sa_column=Column("acct", Text, ForeignKey("accounts.id")))
    category_id: Optional[str] = Field(default=None, sa_column=Column("category", Text, ForeignKey("categories.id")))
    amount: Optional[int] = Field(default=None, sa_column=Column("amount", Integer))
    payee_id: Optional[str] = Field(default=None, sa_column=Column("description", Text, ForeignKey("payees.id")))
    notes: Optional[str] = Field(default=None, sa_column=Column("notes", Text))
    date: Optional[int] = Field(default=None, sa_column=Column("date", Integer))
    financial_id: Optional[str] = Field(default=None, sa_column=Column("financial_id", Text))
    type: Optional[str] = Field(default=None, sa_column=Column("type", Text))
    location: Optional[str] = Field(default=None, sa_column=Column("location", Text))
    error: Optional[str] = Field(default=None, sa_column=Column("error", Text))
    imported_description: Optional[str] = Field(default=None, sa_column=Column("imported_description", Text))
    starting_balance_flag: Optional[int] = Field(
        default=None,
        sa_column=Column("starting_balance_flag", Integer, server_default=text("0")),
    )
    transferred_id: Optional[str] = Field(
        default=None, sa_column=Column("transferred_id", Text, ForeignKey("transactions.id"))
    )
    sort_order: Optional[float] = Field(default=None, sa_column=Column("sort_order", Float))
    tombstone: Optional[int] = Field(default=None, sa_column=Column("tombstone", Integer, server_default=text("0")))
    cleared: Optional[int] = Field(default=None, sa_column=Column("cleared", Integer, server_default=text("1")))
    pending: Optional[int] = Field(default=None, sa_column=Column("pending", Integer, server_default=text("0")))
    parent_id: Optional[str] = Field(default=None, sa_column=Column("parent_id", Text, ForeignKey("transactions.id")))
    schedule_id: Optional[str] = Field(default=None, sa_column=Column("schedule", Text, ForeignKey("schedules.id")))
    reconciled: Optional[int] = Field(default=None, sa_column=Column("reconciled", Integer, server_default=text("0")))
    raw_synced_data: Optional[str] = Field(default=None, sa_column=Column("raw_synced_data", Text))

    account: "Accounts" = Relationship(back_populates="transactions")
    category: Optional["Categories"] = Relationship(
        back_populates="transactions",
        sa_relationship_kwargs={
            "primaryjoin": "and_(Transactions.category_id == Categories.id, Categories.tombstone==0)"
        },
    )
    payee: Optional["Payees"] = Relationship(
        back_populates="transactions",
        sa_relationship_kwargs={"primaryjoin": "and_(Transactions.payee_id == Payees.id, Payees.tombstone==0)"},
    )
    schedule: Optional["Schedules"] = Relationship(
        back_populates="transactions",
        sa_relationship_kwargs={
            "primaryjoin": "and_(Transactions.schedule_id == Schedules.id, Schedules.tombstone==0)"
        },
    )
    parent: Optional["Transactions"] = Relationship(
        back_populates="splits",
        sa_relationship_kwargs={"remote_side": "Transactions.id", "foreign_keys": "Transactions.parent_id"},
    )
    splits: List["Transactions"] = Relationship(
        back_populates="parent",
        sa_relationship_kwargs={
            "primaryjoin": "and_(Transactions.id == remote(Transactions.parent_id), remote(Transactions.tombstone)==0)",
            "order_by": "remote(Transactions.sort_order.desc())",
        },
    )
    transfer: Optional["Transactions"] = Relationship(
        sa_relationship_kwargs={
            "primaryjoin": "and_(Transactions.transferred_id == remote(Transactions.id), remote(Transactions.tombstone)==0)",
            "foreign_keys": "Transactions.transferred_id",
        }
    )

    def get_date(self) -> datetime.date:
        """Returns the transaction date as a datetime.date object, instead of as a string."""
        return int_to_date(self.date)

    def set_date(self, date: datetime.date):
        """Sets the transaction date as a datetime.date object, instead of as a string."""
        self.date = date_to_int(date)

    def set_amount(self, amount: Union[decimal.Decimal, int, float]):
        """Sets the amount as a decimal.Decimal object, instead of as an integer representing the number of cents."""
        self.amount = decimal_to_cents(amount)

    def get_amount(self) -> decimal.Decimal:
        """Returns the amount as a decimal.Decimal, instead of as an integer representing the number of cents."""
        return cents_to_decimal(self.amount)


class ZeroBudgetMonths(SQLModel, table=True):
    __tablename__ = "zero_budget_months"

    id: Optional[str] = Field(default=None, sa_column=Column("id", Text, primary_key=True))
    buffered: Optional[int] = Field(default=None, sa_column=Column("buffered", Integer, server_default=text("0")))


class BaseBudgets(BaseModel):
    """
    Hosts the shared code between both [ZeroBudgets][actual.database.ZeroBudgets] and
    [ReflectBudgets][actual.database.ReflectBudgets].

    The budget will represent a certain month in a certain category.
    """

    id: Optional[str] = Field(default=None, sa_column=Column("id", Text, primary_key=True))
    month: Optional[int] = Field(default=None, sa_column=Column("month", Integer))
    category_id: Optional[str] = Field(default=None, sa_column=Column("category", Text))
    amount: Optional[int] = Field(default=None, sa_column=Column("amount", Integer, server_default=text("0")))

    def get_date(self) -> datetime.date:
        """Returns the transaction date as a datetime.date object, instead of as a string."""
        return int_to_date(self.month, month_only=True)

    def set_date(self, date: datetime.date):
        """
        Sets the transaction date as a datetime.date object, instead of as a string.

        If the date value contains a day, it will be truncated and only the month and year will be inserted, as the
        budget applies to a month.
        """
        self.month = date_to_int(date, month_only=True)

    def set_amount(self, amount: Union[decimal.Decimal, int, float]):
        """Sets the amount as a decimal.Decimal object, instead of as an integer representing the number of cents."""
        self.amount = decimal_to_cents(amount)

    def get_amount(self) -> decimal.Decimal:
        """Returns the amount as a decimal.Decimal, instead of as an integer representing the number of cents."""
        return cents_to_decimal(self.amount)

    @property
    def range(self) -> Tuple[datetime.date, datetime.date]:
        """
        Range of the budget as a tuple [start, end).

        The end date is not inclusive, as it represents the start of the next month.
        """
        return month_range(self.get_date())

    @property
    def balance(self) -> decimal.Decimal:
        """
        Returns the current **spent** balance of the budget.

        The evaluation will take into account the budget month and only selected transactions for the combination month
        and category. Deleted transactions are ignored.

        If you want to get the balance from the frontend, take a look at the query
        [get_accumulated_budgeted_balance][actual.queries.get_accumulated_budgeted_balance] instead.
        """
        budget_start, budget_end = (date_to_int(d) for d in self.range)
        value = object_session(self).scalar(
            select(func.coalesce(func.sum(Transactions.amount), 0)).where(
                Transactions.category_id == self.category_id,
                Transactions.date >= budget_start,
                Transactions.date < budget_end,
                Transactions.is_parent == 0,
                Transactions.tombstone == 0,
            )
        )
        return cents_to_decimal(value)


class ReflectBudgets(BaseBudgets, table=True):
    """
    Stores the budgets, when using tracking budget.

    This table will only contain data for the entries which are created. If a combination of category and budget month
    is not existing, it is assumed that the budget is 0.
    """

    __tablename__ = "reflect_budgets"

    id: Optional[str] = Field(default=None, sa_column=Column("id", Text, primary_key=True))
    month: Optional[int] = Field(default=None, sa_column=Column("month", Integer))
    category_id: Optional[str] = Field(default=None, sa_column=Column("category", ForeignKey("categories.id")))
    amount: Optional[int] = Field(default=None, sa_column=Column("amount", Integer, server_default=text("0")))
    carryover: Optional[int] = Field(default=None, sa_column=Column("carryover", Integer, server_default=text("0")))
    goal: Optional[int] = Field(default=None, sa_column=Column("goal", Integer, server_default=text("null")))
    long_goal: Optional[int] = Field(default=None, sa_column=Column("long_goal", Integer, server_default=text("null")))

    category: "Categories" = Relationship(
        back_populates="reflect_budgets",
        sa_relationship_kwargs={
            "uselist": False,
            "primaryjoin": "and_(ReflectBudgets.category_id == Categories.id, Categories.tombstone == 0)",
        },
    )


class ZeroBudgets(BaseBudgets, table=True):
    """
    Stores the budgets, when using envelope budget (default).

    This table will only contain data for the entries which are created. If a combination of category and budget month
    is not existing, it is assumed that the budget is 0.
    """

    __tablename__ = "zero_budgets"

    id: Optional[str] = Field(default=None, sa_column=Column("id", Text, primary_key=True))
    month: Optional[int] = Field(default=None, sa_column=Column("month", Integer))
    category_id: Optional[str] = Field(default=None, sa_column=Column("category", ForeignKey("categories.id")))
    amount: Optional[int] = Field(default=None, sa_column=Column("amount", Integer, server_default=text("0")))
    carryover: Optional[int] = Field(default=None, sa_column=Column("carryover", Integer, server_default=text("0")))
    goal: Optional[int] = Field(default=None, sa_column=Column("goal", Integer, server_default=text("null")))
    long_goal: Optional[int] = Field(default=None, sa_column=Column("long_goal", Integer, server_default=text("null")))

    category: "Categories" = Relationship(
        back_populates="zero_budgets",
        sa_relationship_kwargs={
            "uselist": False,
            "primaryjoin": "and_(ZeroBudgets.category_id == Categories.id, Categories.tombstone == 0)",
        },
    )


class PendingTransactions(SQLModel, table=True):
    __tablename__ = "pending_transactions"

    id: Optional[str] = Field(default=None, sa_column=Column("id", Text, primary_key=True))
    acct: Optional[int] = Field(default=None, sa_column=Column("acct", ForeignKey("accounts.id")))
    amount: Optional[int] = Field(default=None, sa_column=Column("amount", Integer))
    description: Optional[str] = Field(default=None, sa_column=Column("description", Text))
    date: Optional[str] = Field(default=None, sa_column=Column("date", Text))


for t_entry in SQLModel._sa_registry.mappers:
    t_name = t_entry.entity.__tablename__
    if t_name not in __TABLE_COLUMNS_MAP__:
        __TABLE_COLUMNS_MAP__[t_name] = {"entity": t_entry.entity, "columns": {}, "rev_columns": {}}
    table_columns = list(c.name for c in t_entry.columns)
    # the name and property name of the pydantic property and database column can be different
    for t_key, t_column in dict(t_entry.entity.__dict__).items():
        if hasattr(t_column, "name") and getattr(t_column, "name") in table_columns:
            __TABLE_COLUMNS_MAP__[t_name]["columns"][t_column.name] = t_key
            # add also rever columns mappings for reverse mapping
            __TABLE_COLUMNS_MAP__[t_name]["rev_columns"][t_key] = t_column.name



import datetime
import uuid

import pytest

from actual import ActualError
from actual.exceptions import ActualSplitTransactionError
from actual.queries import (
    create_account,
    create_category,
    create_payee,
    create_transaction,
)
from actual.rules import (
    Action,
    ActionType,
    Condition,
    ConditionType,
    Rule,
    RuleSet,
    ValueType,
    condition_evaluation,
)


def test_category_rule(session):
    # create basic items
    acct = create_account(session, "Bank")
    cat = create_category(session, "Food", "Expenses")
    payee = create_payee(session, "My payee")
    # create rule
    condition = Condition(field="category", op="is", value=cat)
    action = Action(field="description", value=payee)
    rule = Rule(conditions=[condition], actions=[action], operation="all")
    rs = RuleSet(rules=[])
    assert list(rs) == []
    rs.add(rule)
    # run for one transaction
    t = create_transaction(session, datetime.date(2024, 1, 1), acct, "", category=cat)
    rs.run(t)
    # evaluate if things match
    assert t.payee_id == payee.id
    assert (
        str(rs) == f"If all of these conditions match 'category' is '{cat.id}' "
        f"then set 'description' to '{payee.id}'"
    )
    # check if it ignores the input when making the category None
    t.category_id = None
    assert condition.run(t) is False


def test_datetime_rule(session):
    acct = create_account(session, "Bank")
    t = create_transaction(session, datetime.date(2024, 1, 1), acct, "")
    condition = Condition(field="date", op="isapprox", value=datetime.date(2024, 1, 2))
    action = Action(field="date", value="2024-01-30")
    rs = RuleSet(rules=[Rule(conditions=[condition], actions=[action], operation="any")])
    # run only first stage
    rs.run([t], stage=None)
    target_date = datetime.date(2024, 1, 30)
    assert t.get_date() == target_date
    # try the is not
    assert Condition(field="date", op="is", value=target_date).run(t) is True
    # try the comparison operators
    assert Condition(field="date", op="gte", value=target_date).run(t) is True
    assert Condition(field="date", op="lte", value=target_date).run(t) is True
    assert Condition(field="date", op="gte", value=target_date - datetime.timedelta(days=1)).run(t) is True
    assert Condition(field="date", op="lte", value=target_date + datetime.timedelta(days=1)).run(t) is True
    assert Condition(field="date", op="lt", value=target_date).run(t) is False
    assert Condition(field="date", op="gt", value=target_date).run(t) is False
    assert Condition(field="date", op="gt", value=target_date - datetime.timedelta(days=1)).run(t) is True
    assert Condition(field="date", op="lt", value=target_date + datetime.timedelta(days=1)).run(t) is True


def test_string_condition(session):
    acct = create_account(session, "Bank")
    t = create_transaction(session, datetime.date(2024, 1, 1), acct, "", "foo")
    assert Condition(field="notes", op="oneOf", value=["foo", "bar"]).run(t) is True
    assert Condition(field="notes", op="notOneOf", value=["foo", "bar"]).run(t) is False
    assert Condition(field="notes", op="contains", value="fo").run(t) is True
    assert Condition(field="notes", op="contains", value="foobar").run(t) is False
    assert Condition(field="notes", op="matches", value="f.*").run(t) is True
    assert Condition(field="notes", op="matches", value="g.*").run(t) is False
    assert Condition(field="notes", op="doesNotContain", value="foo").run(t) is False
    assert Condition(field="notes", op="doesNotContain", value="foobar").run(t) is True
    # case insensitive entries
    assert Condition(field="notes", op="oneOf", value=["FOO", "BAR"]).run(t) is True
    assert Condition(field="notes", op="notOneOf", value=["FOO", "BAR"]).run(t) is False
    assert Condition(field="notes", op="contains", value="FO").run(t) is True
    assert Condition(field="notes", op="contains", value="FOOBAR").run(t) is False
    assert Condition(field="notes", op="matches", value="F.*").run(t) is True
    assert Condition(field="notes", op="matches", value="G.*").run(t) is False
    assert Condition(field="notes", op="doesNotContain", value="FOO").run(t) is False
    assert Condition(field="notes", op="doesNotContain", value="FOOBAR").run(t) is True


def test_has_tags(session):
    acct = create_account(session, "Bank")
    t = create_transaction(session, datetime.date(2024, 1, 1), acct, "", "foo #bar #✨ #🙂‍↔️")
    assert Condition(field="notes", op="hasTags", value="#bar").run(t) is True
    assert Condition(field="notes", op="hasTags", value="#foo").run(t) is False
    # test other unicode entries
    assert Condition(field="notes", op="hasTags", value="#emoji #✨").run(t) is True
    assert Condition(field="notes", op="hasTags", value="#🙂‍↔️").run(t) is True  # new emojis should be supported
    assert Condition(field="notes", op="hasTags", value="bar").run(t) is False  # individual string will not match


@pytest.mark.parametrize(
    "op,condition_value,value,expected_result",
    [
        ("contains", "supermarket", "Best Supermarket", True),
        ("contains", "supermarket", None, False),
        ("oneOf", ["my supermarket", "other supermarket"], "MY SUPERMARKET", True),
        ("oneOf", ["supermarket"], None, False),
        ("matches", "market", "hypermarket", True),
    ],
)
def test_imported_payee_condition(session, op, condition_value, value, expected_result):
    create_account(session, "Bank")
    t = create_transaction(session, datetime.date(2024, 1, 1), "Bank", "", amount=5, imported_payee=value)
    condition = {"field": "imported_description", "type": "imported_payee", "op": op, "value": condition_value}
    cond = Condition.model_validate(condition)
    assert cond.run(t) == expected_result


def test_numeric_condition(session):
    create_account(session, "Bank")
    t = create_transaction(session, datetime.date(2024, 1, 1), "Bank", "", amount=5)
    c1 = Condition(field="amount_inflow", op="gt", value=10.0)
    assert "inflow" in c1.options
    assert c1.run(t) is False
    c2 = Condition(field="amount_outflow", op="lt", value=-10.0)
    assert "outflow" in c2.options
    assert c2.run(t) is False  # outflow, so the comparison should be with the positive value
    # isapprox condition
    c2 = Condition(field="amount", op="isapprox", value=5.1)
    assert c2.run(t) is True
    c3 = Condition(field="amount", op="isapprox", value=5.5)
    assert c3.run(t) is False
    # isbetween condition
    c4 = Condition(field="amount", op="isbetween", value={"num1": 5.0, "num2": 10.0})
    assert c4.run(t) is True
    assert str(c4) == "'amount' isbetween (500, 1000)"  # value gets converted when input as float


def test_complex_rule(session):
    # create basic items
    acct = create_account(session, "Bank")
    cat = create_category(session, "Food", "Expenses")
    cat_extra = create_category(session, "Restaurants", "Expenses")
    payee = create_payee(session, "My payee")
    # create rule set
    rs = RuleSet(
        rules=[
            Rule(
                conditions=[
                    Condition(
                        field="category",
                        op=ConditionType.ONE_OF,
                        value=[cat],
                    ),
                    Condition(
                        field="category",
                        op=ConditionType.NOT_ONE_OF,
                        value=[cat_extra],
                    ),
                    Condition(field="description", op="isNot", value=str(uuid.uuid4())),  # should not match
                ],
                actions=[Action(field="cleared", value=True)],
                operation="all",
            )
        ]
    )
    t_true = create_transaction(session, datetime.date(2024, 1, 1), acct, payee, category=cat)
    t_false = create_transaction(session, datetime.date(2024, 1, 1), acct, payee)
    rs.run([t_true, t_false])
    assert t_true.cleared == 1
    assert t_false.cleared == 0


def test_invalid_inputs():
    with pytest.raises(ValueError):
        Condition(field="amount", op="gt", value="foo")
    with pytest.raises(ValueError):
        Condition(field="amount", op="contains", value=10)
    with pytest.raises(ValueError):
        Action(field="date", value="foo")
    with pytest.raises(ValueError):
        Condition(field="description", op="is", value="foo")  # not an uuid
    with pytest.raises(ValueError):
        Condition(field="amount", op="isbetween", value=5)
    with pytest.raises(ActualError):
        Action(field="notes", op="set-split-amount", value="foo").run(None)  # noqa: use None instead of transaction
    with pytest.raises(ActualError):
        condition_evaluation(None, "foo", "foo")  # noqa: use None instead of transaction
    assert Condition(field="notes", op="is", value=None).get_value() is None  # noqa: handle when value is None


def test_value_type_condition_validation():
    assert ValueType.DATE.is_valid(ConditionType.IS_APPROX) is True
    assert ValueType.DATE.is_valid(ConditionType.CONTAINS) is False
    assert ValueType.NUMBER.is_valid(ConditionType.IS_BETWEEN) is True
    assert ValueType.NUMBER.is_valid(ConditionType.DOES_NOT_CONTAIN) is False
    assert ValueType.BOOLEAN.is_valid(ConditionType.IS) is True
    assert ValueType.ID.is_valid(ConditionType.NOT_ONE_OF) is True
    assert ValueType.ID.is_valid(ConditionType.CONTAINS) is False
    assert ValueType.STRING.is_valid(ConditionType.CONTAINS) is True
    assert ValueType.STRING.is_valid(ConditionType.GT) is False
    assert ValueType.IMPORTED_PAYEE.is_valid(ConditionType.CONTAINS) is True
    assert ValueType.IMPORTED_PAYEE.is_valid(ConditionType.GT) is False


def test_value_type_value_validation():
    assert ValueType.DATE.validate(20241004) is True
    assert ValueType.DATE.validate(123) is False
    assert ValueType.DATE.validate("2024-10-04") is True
    assert ValueType.STRING.validate("") is True
    assert ValueType.STRING.validate(123) is False
    assert ValueType.NUMBER.validate(123) is True
    assert ValueType.NUMBER.validate(1.23) is False  # noqa: test just in case
    assert ValueType.NUMBER.validate("123") is False
    assert ValueType.ID.validate("1c1a1707-15ea-4051-b98a-e400ee2900c7") is True
    assert ValueType.ID.validate("foo") is False
    assert ValueType.BOOLEAN.validate(True) is True
    assert ValueType.BOOLEAN.validate("") is False
    assert ValueType.IMPORTED_PAYEE.validate("") is True
    assert ValueType.IMPORTED_PAYEE.validate(1) is False
    # list and NoneType
    assert ValueType.DATE.validate(None)
    assert ValueType.DATE.validate(["2024-10-04"], ConditionType.ONE_OF) is True


def test_value_type_from_field():
    assert ValueType.from_field("description") == ValueType.ID
    assert ValueType.from_field("amount") == ValueType.NUMBER
    assert ValueType.from_field("notes") == ValueType.STRING
    assert ValueType.from_field("date") == ValueType.DATE
    assert ValueType.from_field("cleared") == ValueType.BOOLEAN
    assert ValueType.from_field("imported_description") == ValueType.IMPORTED_PAYEE
    with pytest.raises(ValueError):
        ValueType.from_field("foo")


@pytest.mark.parametrize(
    "method,value,expected_splits",
    [
        ("remainder", None, [0.50, 4.50]),
        ("fixed-amount", 100, [0.40, 1.00, 3.60]),
        ("fixed-percent", 20, [0.50, 1.00, 3.50]),
    ],
)
def test_set_split_amount(session, method, value, expected_splits):
    acct = create_account(session, "Bank")
    cat = create_category(session, "Food", "Expenses")
    payee = create_payee(session, "My payee")
    alternative_payee = create_payee(session, "My other payee")

    rs = RuleSet(
        rules=[
            Rule(
                conditions=[Condition(field="category", op=ConditionType.ONE_OF, value=[cat])],
                actions=[
                    Action(
                        field=None,
                        op=ActionType.SET_SPLIT_AMOUNT,
                        value=10,
                        options={"splitIndex": 1, "method": "fixed-percent"},
                    ),
                    Action(
                        field=None,
                        op=ActionType.SET_SPLIT_AMOUNT,
                        value=value,
                        options={"splitIndex": 2, "method": method},
                    ),
                    # add one action that changes the second split payee
                    Action(
                        field="description", op=ActionType.SET, value=alternative_payee.id, options={"splitIndex": 2}
                    ),
                ],
            )
        ]
    )
    t = create_transaction(session, datetime.date(2024, 1, 1), acct, payee, category=cat, amount=5.0)
    session.flush()
    rs.run(t)
    session.refresh(t)
    assert [float(s.get_amount()) for s in t.splits] == expected_splits
    # check the first split has the original payee, and the second split has the payee from the action
    assert t.splits[0].payee_id == payee.id
    assert t.splits[1].payee_id == alternative_payee.id
    # check string comparison
    assert (
        str(rs.rules[0]) == f"If all of these conditions match 'category' oneOf ['{cat.id}'] then "
        f"allocate a fixed-percent at Split 1: 10, "
        f"allocate a {method} at Split 2: {value}, "
        f"set 'description' at Split 2 to '{alternative_payee.id}'"
    )


@pytest.mark.parametrize(
    "method,n,expected_splits",
    [
        # test equal remainders
        ("remainder", 1, [5.00]),
        ("remainder", 2, [2.50, 2.50]),
        ("remainder", 3, [1.67, 1.67, 1.66]),
        ("remainder", 4, [1.25, 1.25, 1.25, 1.25]),
        ("remainder", 5, [1.00, 1.00, 1.00, 1.00, 1.00]),
        ("remainder", 6, [0.83, 0.83, 0.83, 0.83, 0.83, 0.85]),
        # and fixed amount
        ("fixed-amount", 1, [1.0, 4.0]),
        ("fixed-amount", 2, [1.0, 1.0, 3.0]),
        ("fixed-amount", 3, [1.0, 1.0, 1.0, 2.0]),
        ("fixed-amount", 4, [1.0, 1.0, 1.0, 1.0, 1.0]),
        ("fixed-amount", 5, [1.0, 1.0, 1.0, 1.0, 1.0]),
        ("fixed-amount", 6, [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, -1.0]),
    ],
)
def test_split_amount_equal_parts(session, method, n, expected_splits):
    acct = create_account(session, "Bank")
    actions = [
        Action(
            field=None,
            op=ActionType.SET_SPLIT_AMOUNT,
            value=100,  # value is only used for fixed-amount
            options={"splitIndex": i + 1, "method": method},
        )
        for i in range(n)
    ]
    rs = Rule(conditions=[], actions=actions)
    t = create_transaction(session, datetime.date(2024, 1, 1), acct, "", amount=5.0)
    session.flush()
    # test split amounts
    splits = rs.set_split_amount(t)
    assert [float(s.get_amount()) for s in splits] == expected_splits


def test_set_split_amount_exception(session, mocker):
    mocker.patch("actual.rules.sum", lambda x: 0)

    acct = create_account(session, "Bank")
    cat = create_category(session, "Food", "Expenses")
    payee = create_payee(session, "My payee")

    rs = RuleSet(
        rules=[
            Rule(
                conditions=[Condition(field="category", op=ConditionType.ONE_OF, value=[cat])],
                actions=[
                    Action(
                        field=None,
                        op=ActionType.SET_SPLIT_AMOUNT,
                        value=10,
                        options={"splitIndex": 1, "method": "fixed-percent"},
                    )
                ],
            )
        ]
    )
    t = create_transaction(session, datetime.date(2024, 1, 1), acct, payee, category=cat, amount=5.0)
    session.flush()
    with pytest.raises(ActualSplitTransactionError):
        rs.run(t)


@pytest.mark.parametrize(
    "operation,value,note,expected",
    [
        ("append-notes", "bar", "foo", "foobar"),
        ("prepend-notes", "bar", "foo", "barfoo"),
        ("append-notes", "bar", None, "bar"),
        ("prepend-notes", "bar", None, "bar"),
    ],
)
def test_preppend_append_notes(session, operation, value, note, expected):
    create_account(session, "Bank")
    t = create_transaction(session, datetime.date(2024, 1, 1), "Bank", "", notes=note)
    action = Action(field="description", op=operation, value=value)
    action.run(t)
    assert t.notes == expected
    action.run(t)  # second iteration should not update the result
    assert t.notes == expected
    assert f"{operation.split('-')[0]} to notes '{value}'" in str(action)


def test_set_transfer_payee_rule(session):
    bank = create_account(session, "Bank")
    t = create_transaction(session, datetime.date(2024, 1, 1), "Bank", amount=10)
    action = Action(field="description", op="set", value=bank.payee.id)
    action.run(t)
    assert t.transferred_id is not None
    assert t.transfer is not None
    assert t.get_amount() == -t.transfer.get_amount()
    assert t.id == t.transfer.transferred_id
    assert t.transferred_id == t.transfer.id
